/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package ca.mcscert.jtet.cvc3generator;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import ca.mcscert.jtet.expression.*;
import org.apache.commons.lang3.StringUtils;

/**
 *
 * @author matthew
 */
final public class CVC3ExpressionGenerator implements ExpressionGenerator {
    public final static CVC3ExpressionGenerator Generator = new CVC3ExpressionGenerator();

    // For operations that don't have specific signed/unsigned representations, use this function.
    private static String GetIndifferentFixedPointFunctionFor(BinaryOperation op) {
        switch (op) {
            case Addition:
                return "BVPLUS(";
            case Minus:
                return "BVSUB(";
            case Multiplication:
                return "BVMULT(";
        }
        throw new IllegalArgumentException("Operation " + op + " doesn't have a fixed point function!");
    }

    private static String GetSignedFixedPointFunctionFor(BinaryOperation op) {
        switch (op) {
            case Division:
                return "BVSDIV(";
            case GreaterThen:
                return "BVSGT(";
            case GreaterThenEqual:
                return "BVSGE(";
            case LessThen:
                return "BVSLT(";
            case LessThenEqual:
                return "BVSLE(";
        }
        return GetIndifferentFixedPointFunctionFor(op);
    }

    private static String GetUnsignedFixedPointFunctionFor(BinaryOperation op) {
        switch (op) {
            case Division:
                return "BVUDIV(";
            case GreaterThen:
                return "BVGT(";
            case GreaterThenEqual:
                return "BVGE(";
            case LessThen:
                return "BVLT(";
            case LessThenEqual:
                return "BVLE(";
        }
        return GetIndifferentFixedPointFunctionFor(op);
    }

    private static String GetInfixSymbolFor(BinaryOperation op) {
        switch (op) {
            case Addition:
                return "+";
            case Minus:
                return "-";
            case Multiplication:
                return "*";
            case Division:
                return "/";
            case Exponentiation:
                return "^";
            case GreaterThen:
                return ">";
            case GreaterThenEqual:
                return ">=";
            case LessThen:
                return "<";
            case LessThenEqual:
                return "<=";
            case Equals:
                return "=";
            case NotEquals:
                return "/=";
            case BooleanAnd:
                return "AND";
            case BooleanOr:
                return "OR";
            case Implies:
                return "=>";
        }
        throw new RuntimeException("Should never happen!");
    }

    public static String ConvertToOutputOp(BinaryOperation op, Type type) {
        switch (OpTypeForOpAndType(op, type)) {
            case Function:
                String func = GetSignedFixedPointFunctionFor(op);
                if (!((FixedPointType) type).isSigned()) {
                    func = GetUnsignedFixedPointFunctionFor(op);
                }
                if (op == BinaryOperation.Addition || op == BinaryOperation.Minus || op == BinaryOperation.Multiplication) {
                    FixedPointType typeF = (FixedPointType) type;
                    return func + typeF.digits() + ",";
                } else {
                    return func;
                }
            case CenterOp:
                return GetInfixSymbolFor(op);
            default:
                throw new IllegalStateException("CVC3 Operation type asked for that doesn't exist!");
        }
    }

    private static OperationType OpTypeForOpAndType(BinaryOperation op, Type type) {
        if (type instanceof FixedPointType && op != BinaryOperation.Equals && op != BinaryOperation.NotEquals) {
            if (op == BinaryOperation.Exponentiation) {
                throw new IllegalArgumentException("Cannot generate exponential functions in fixed point!");
            }
            return OperationType.Function;
        } else {
            return OperationType.CenterOp;
        }
    }

    @Override
    public String GenerateBinaryOperation(BinaryOperation op, String lhsExp, String rhsExp, Type inputType) {
        switch (CVC3ExpressionGenerator.OpTypeForOpAndType(op, inputType)) {
            case CenterOp:
                return "(" + lhsExp + " " + CVC3ExpressionGenerator.ConvertToOutputOp(op, inputType) + " " + rhsExp + ")";
            case Function:
                return CVC3ExpressionGenerator.ConvertToOutputOp(op, inputType) + lhsExp + "," + rhsExp + ")";
            default:
                throw new IllegalStateException("Cannot output expression operation for requested operation type.");
        }
    }

    private static String GetUnaryOperationSymbolFor(UnaryOperation op) {
        switch (op) {
            case Negation:
                return "NOT";
            case Minus:
                return "-";
        }
        throw new RuntimeException("Should never happen!  Asked for unhandled op: " + op + "!");
    }

    @Override
    public String GenerateUnaryOperation(UnaryOperation op, String expression, Type inputType) {
        switch (CVC3ExpressionGenerator.OpTypeForOpAndType(op, inputType)) {
            case PrefixOp:
                return "(" + CVC3ExpressionGenerator.ConvertToOutputOp(op, inputType) + " " + expression + ")";
            case Function:
                return CVC3ExpressionGenerator.ConvertToOutputOp(op, inputType) + "(" + expression + ")";
            default:
                throw new IllegalStateException("Cannot output expression operation for requested operation type.");
        }
    }

    private static OperationType OpTypeForOpAndType(UnaryOperation op, Type usedType) {
        if (usedType instanceof FixedPointType) {
            return OperationType.Function;
        }
        return OperationType.PrefixOp;
    }

    private static String ConvertToOutputOp(UnaryOperation op, Type usedType) {
        switch (op) {
            case Negation:
                return "NOT";
            case Minus:
                if (usedType instanceof FixedPointType) {
                    return "BVUMINUS";
                } else {
                    return "-";
                }
        }
        throw new IllegalArgumentException("Unhandled operation: " + op);
    }

    @Override
    public String GenerateLiteralValue(String value, Type requestedType) {
        if (requestedType instanceof RealType) {
            if (value.contains(".")) {
                String[] fraction = value.split("\\.");

                String res = "(";
                res += StringUtils.join(fraction, "");
                res += "/1" + StringUtils.repeat('0', fraction[1].length()) + ")";

                return res;
            } else {
                return value;
            }
        } else if (requestedType instanceof FixedPointType) {
            FixedPointType type = (FixedPointType) requestedType;
            BigDecimal decimal = new BigDecimal(value);
            decimal = decimal.multiply(BigDecimal.valueOf(1 << type.fractionPart()));
            BigInteger intOut = decimal.toBigIntegerExact();

            final StringBuilder output = new StringBuilder(type.digits());
            output.append("0bin");

            if (intOut.signum() < 0) {
                output.append(StringUtils.repeat('1', type.digits() - intOut.bitLength()));
            } else {
                output.append(StringUtils.repeat('0', type.digits() - intOut.bitLength()));
            }

            //This gets the two's compliment representation.  Ugly for loop, but best I could find.
            for (int i = intOut.bitLength() - 1; i >= 0; --i) {
                output.append((intOut.testBit(i)) ? '1' : '0');
            }
            return output.toString();
        } else {
            throw new IllegalArgumentException("Can't cast literal to that type!");
        }
    }
}
