/*
 * Copyright (C) 2014 Yanjun Jiang <jiangy76@mcmaster.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package ca.mcscert.jtet.eventbgenerator;

import ca.mcscert.jtet.expression.*;

import java.util.Collection;

/**
 *
 * @author Yanjun Jiang <jiangy76@mcmaster.ca>
 */
final public class EventBVariablesDeclarationGenerator implements VariablesDeclarationGenerator {
    public final static EventBVariablesDeclarationGenerator Generator = new EventBVariablesDeclarationGenerator();

    private String GenerateType(Type type) {
        if (type instanceof RealType) {
            return "ℤ";
        } else {
            throw new IllegalArgumentException("EventB Doesn't handle this variable type (Class: " + type.getClass().getName() + ")");
        }
    }

    @Override
    public String GenerateVariablesDeclaration(Collection<Variable> variables) {
        final StringBuilder variablesRelatedXml = new StringBuilder();

        generateVariablesXml(variablesRelatedXml, variables);
        generateInvariantsXml(variablesRelatedXml, variables);
        generateInitEventXml(variablesRelatedXml, variables);

        return variablesRelatedXml.toString();
    }

    private void generateVariablesXml(StringBuilder ret, Collection<Variable> variables) {
        final String variableXmlPost = "\"/>\n";

        //generate all the variables one by one.
        for(Variable var : variables) {
            generateVariableXmlPre(ret, var.outputName());
            ret.append(var.outputName());
            ret.append(variableXmlPost);
        }
    }

    private void generateInvariantXmlPre(StringBuilder ret, int invariantNum) {
        ret.append("<org.eventb.core.invariant name=\"Invariant")
                .append(invariantNum)
                .append("\" org.eventb.core.label=\"inv")
                .append(invariantNum)
                .append("\" org.eventb.core.predicate=\"");
    }

    private void generateVariableXmlPre(StringBuilder ret, String variableName) {
        ret.append("<org.eventb.core.variable de.prob.units.inferredUnitPragmaAttribute=\"\" de.prob.units.unitPragmaAttribute=\"\" name=\"")
                .append(variableName)
                .append("\" org.eventb.core.identifier=\"");
    }

    private void generateInvariantsXml(StringBuilder ret, Collection<Variable> variables) {
        int  invariantsNum = 1;
        final String invariantXmlPost = "\"/>\n";

        //generate all the Invariants one by one.
        for (Variable var : variables) {
            generateInvariantXmlPre(ret, invariantsNum);
            ret.append(var.outputName());
            ret.append(" ∈ ");
            ret.append(GenerateType(var.type()));
            ret.append(invariantXmlPost);
            invariantsNum++;
        }
    }

    private String generateInitEventXmlPre() {
        return "<org.eventb.core.event name=\"Event0\" org.eventb.core.convergence=\"0\" org.eventb.core.extended=\"false\" org.eventb.core.label=\"INITIALISATION\">\n";
    }

    private String generateInitEventXmlPost() {
        return "</org.eventb.core.event>\n";
    }

    private void generateInitEventActionXmlPre(StringBuilder ret, int actionNo) {
        ret.append("<org.eventb.core.action name=\"InitAction")
                .append(actionNo)
                .append("\" org.eventb.core.assignment=\"");
    }

    private void generateInitEventActionXmlPost(StringBuilder ret, int actionNo) {
        ret.append("\" org.eventb.core.label=\"act")
                .append(actionNo)
                .append("\"/>");
    }

    private void generateInitEventXml(StringBuilder ret, Collection<Variable> variables) {
        ret.append(generateInitEventXmlPre());
        generateInitEventActionsXml(ret, variables);
        ret.append(generateInitEventXmlPost());
    }

    private void generateInitEventActionsXml(StringBuilder ret, Collection<Variable> variables) {

        //generate all the Initialisation actions one by one.
        int actionNo = 0;
        for (Variable var : variables) {
            actionNo++;
            generateInitEventActionXmlPre(ret, actionNo);
            ret.append(var.outputName());
            ret.append(" :∈ ");
            ret.append(GenerateType(var.type()));
            generateInitEventActionXmlPost(ret, actionNo);
            ret.append("\n");
        }
    }
}
