/*
 * Copyright (C) 2014 Yanjun Jiang <jiangy76@mcmaster.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.eventbgenerator;

import ca.mcscert.jtet.expression.BooleanType;
import ca.mcscert.jtet.expression.Expression;
import ca.mcscert.jtet.expression.Type;
import ca.mcscert.jtet.expression.Variable;
import ca.mcscert.jtet.tabularexpression.HierarchicalGridDepthFirstCheckerGenerator;
import org.apache.commons.lang3.StringUtils;

import java.util.*;

/**
 * This class generate common non-initialisation events
 * @author Yanjun Jiang <jiangy76@mcmaster.ca>
 */
final public class HierarchicalGridEventBGenerator implements HierarchicalGridDepthFirstCheckerGenerator {

    public HierarchicalGridEventBGenerator(RefinementMode refineMode, int refineLayerNo) {
        m_refineMode = refineMode;
        m_refineLayerNo = refineLayerNo;
    }

    private String generateEventGuardXml(int guardNo, String guard) {
        return "<org.eventb.core.guard name=\"Guard" + guardNo + "\" org.eventb.core.label=\"grd" + + guardNo + "\" org.eventb.core.predicate=\"" + guard +"\"/>\n";
    }

    private void generateOneEventActionXml(StringBuilder ret, int actionNo, String outputVariableName, String outputValue, String assignmentOperator) {
        ret.append("<org.eventb.core.action name=\"Action")
                .append(actionNo)
                .append("\" org.eventb.core.assignment=\"")
                .append(outputVariableName)
                .append(" ")
                .append(assignmentOperator)
                .append(" ")
                .append(outputValue)
                .append("\" org.eventb.core.label=\"act")
                .append(actionNo)
                .append("\"/>\n");
    }

    private String generateEventActionsXml() {
        final StringBuilder ret = new StringBuilder();
        String outputVariableName;
        String outputValue;
        int actionNo = 1;

        for (Map.Entry<Variable, List<String>> variableOutputs : m_collectedOutputValues.entrySet()) {
            List<String> collectedOutputValues = variableOutputs.getValue();
            Variable outputVariable = variableOutputs.getKey();
            outputVariableName = outputVariable.name();
            if (collectedOutputValues.size() == 1) {
                outputValue = collectedOutputValues.get(0);
                generateOneEventActionXml(ret, actionNo, outputVariableName, outputValue, "≔");
            } else {
                outputValue = "{";
                outputValue += StringUtils.join(collectedOutputValues, ',');
                outputValue += "}";
                generateOneEventActionXml(ret, actionNo, outputVariableName, outputValue, ":∈");
            }
            actionNo++;
        }

        return ret.toString();
    }

    private String getEventName(){
        return "Event" + m_eventNo;
    }

    private String getEventLabelName(String eventNo){
        return "evt" + eventNo;
    }

    private String generateEventLabelXml() {
        return "<org.eventb.core.event name=\"" + getEventName() + "\" org.eventb.core.convergence=\"0\" org.eventb.core.extended=\"false\" org.eventb.core.label=\"" + getEventLabelName(m_eventNo) + "\">\n";
    }

    private String getRefineName(){
        if (m_cellLayerNo < m_refineLayerNo || m_parentEventNo.equals("")) {
            return getEventLabelName(m_eventNo);
        } else {
            return getEventLabelName(m_parentEventNo);
        }
    }
    private String generateEventRefineRelationXml(){
        return "<org.eventb.core.refinesEvent name=\"" + getRefineName() + "\" org.eventb.core.target=\"" + getRefineName() + "\"/>\n";
    }

    private String generateEventXml() {
        int guardNo = 1;
        final StringBuilder ret = new StringBuilder();
        final String eventXmlPost = "</org.eventb.core.event>\n";

        ret.append(generateEventLabelXml());

        for (String parentCellContent: m_eventBGuards){
            ret.append(generateEventGuardXml(guardNo++, parentCellContent));
        }

        ret.append(generateEventActionsXml());

        if (isRefineRelationNecessary()) {
            ret.append(generateEventRefineRelationXml());
        }

        ret.append(eventXmlPost);

        m_collectedOutputValues.clear();

        return ret.toString();
    }

    public void generateEventNo() {
        m_eventNo = m_parentEventNo;

        if (!m_eventNo.equals("")) {
            m_eventNo += "_";
            m_eventNo += m_subGridCellIndex;
        }
        else{
            m_eventNo += m_subGridCellIndex;
        }
    }

    @Override
    public void descendIntoGridFromCell(Expression inputCellExpression) {
        m_oldParentEventNo.push(m_parentEventNo);
        m_oldSubGridCellIndex.push(m_subGridCellIndex);
        m_subGridCellIndex = 0;

        m_cellLayerNo++;
        m_parentEventNo = m_eventNo;
        m_eventBGuards.push(handleConditionCellExpression(inputCellExpression));
    }

    @Override
    public void ascendFromGrid() {
        m_cellLayerNo--;
        m_parentEventNo = m_oldParentEventNo.pop();
        m_subGridCellIndex = m_oldSubGridCellIndex.pop();
        if (m_cellLayerNo == m_refineLayerNo) {
            m_eventsXml += generateEventXml();//outputEdgeCells
        }
        m_eventBGuards.pop();
    }

    @Override
    public void handleEdgeCell(Expression inputCellExpression) {
        m_subGridCellIndex++;
        generatorGenerateEventNo();
    }

    @Override
    public void handleLeafCell(Expression inputCellExpression, Map<Variable, Expression> outputCellsExpression) {
        m_subGridCellIndex++;
        generatorGenerateEventNo();

        m_eventBGuards.push(handleConditionCellExpression(inputCellExpression));
        collectEdgeGridValues(outputCellsExpression);
        if (m_cellLayerNo <= m_refineLayerNo) {
            m_eventsXml += generateEventXml();//outputLeafCells
        }
        m_eventBGuards.pop();
    }

    private void generatorGenerateEventNo() {
        if (m_cellLayerNo <= m_refineLayerNo) {
            generateEventNo();
        }
    }

    private void collectEdgeGridValues(Map<Variable, Expression> outputCellsExpression) {
        for(Map.Entry<Variable, Expression> outputCell : outputCellsExpression.entrySet()) {
            Expression outputExpression = outputCell.getValue();
            String convertedValue = convertExpressionToEventB(outputExpression, outputCell.getKey().type());
            if(!m_collectedOutputValues.containsKey(outputCell.getKey())) {
                m_collectedOutputValues.put(outputCell.getKey(), new ArrayList<String>());
            }
            m_collectedOutputValues.get(outputCell.getKey()).add(convertedValue);
        }
    }

    private String convertExpressionToEventB(Expression expression, Type type) {
        return expression.getCheckerOutput(m_eventBExpressionGenerator, type);
    }

    private String handleConditionCellExpression(Expression inputCellExpression) {
        return convertExpressionToEventB(inputCellExpression, BooleanType.Type);
    }

    public boolean isRefineRelationNecessary() {
        return (0 != m_refineLayerNo) && (RefinementMode.WithoutRefinement != m_refineMode);//the first layer also has not refinement.
    }

    @Override
    public String getFinalString() {
        return m_eventsXml;
    }

    int m_cellLayerNo = 0;
    int m_subGridCellIndex = 0;
    int m_refineLayerNo = 0;//is also the current column number
    String m_eventsXml = "";
    String m_eventNo = "";//like evt1_3_2
    String m_parentEventNo = "";//like evt1_3
    Map<Variable, List<String>> m_collectedOutputValues = new LinkedHashMap<Variable, List<String>>();
    EventBExpressionGenerator m_eventBExpressionGenerator = new EventBExpressionGenerator();
    Stack<String> m_eventBGuards = new Stack<String>();
    Stack<String> m_oldParentEventNo = new Stack<String>();
    Stack<Integer> m_oldSubGridCellIndex = new Stack<Integer>();
    RefinementMode m_refineMode;
}
