/*
 * Copyright (C) 2013 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.expression;

/**
 * Provides the visitor interface for the expression generation.
 * <p>
 * Users looking to extract information from the AST can implement this interface.  It is designed to work with generators
 * to produce a new string representing the expression, but can be used to simply visit each node as well.
 * @author Matthew Dawson
 */
public interface ExpressionGenerator {
    /**
     * Visit a node with a unary expression.
     * <p>
     * When a unary expression (like negation) occurs, this function is called passing in the various information realted
     * to it.
     * @param op The unary op at this node.
     * @param expression The string representing the expression this node is applied to.
     * @param inputType The type of the input (which should match the output).
     * @return A string representing this part of the expression.
     */
    public String GenerateUnaryOperation(UnaryOperation op, String expression, Type inputType);

    /**
     * Visit a node with a binary operation.
     * <p>
     * Similar to {@link #GenerateUnaryOperation}, this handles the binary operations.  Binary operations do distinguish
     * between the left and right hand side when calling into this function, and matches what is in the AST.  Also, the
     * input type is always the output type fo the two given expressions.  It is assumed the output type of this expression
     * is either the input type, or is a boolean for comparison operators.  That type change is handled by jTET's type system.
     * @param op The binary operation at this node.
     * @param lhsExp The string representing the left hand side of this operation.
     * @param rhsExp The string representing the right hand side of this operation.
     * @param inputType The type of the input.
     * @return A string representing this part of the expression.
     */
    public String GenerateBinaryOperation(BinaryOperation op, String lhsExp, String rhsExp, Type inputType);

    /**
     * Generates a string representing the given literal.
     * <p>
     * This converts the string form of the given literal into an appropriate output for the language.  Currently this
     * only deals with real numbers.
     * @param value A string representing a given number.
     * @param requestedType The type to use for the output.
     * @return A string representing this literal in an appropriate form.
     */
    public String GenerateLiteralValue(String value, Type requestedType);
}
