/*
 * Copyright (C) 2014 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.expression;

import java.util.*;

/**
 * @author Matthew Dawson <matthew@mjdsystems.ca>
 */
final public class PartialVariableCollection {
    public PartialVariableCollection(Map<String, Variable> variables, Set<EnumerationType> enumerationTypes) {
        m_variables = Collections.unmodifiableMap(variables);
        if (enumerationTypes == null) {
            enumerationTypes = Collections.emptySet();
        }
        m_enumerationTypes = Collections.unmodifiableSet(enumerationTypes);
    }

    /**
     * Construct PartialVariableCollection from a list of variables
     *
     * Often when constructing a PartialVariableCollection, you have a compile time defined list of variables.  This
     * provides a simple method to construct a partial variable collection from such a list.
     * @param variables The variables to store in this collection.
     */
    public PartialVariableCollection(Variable... variables) {
        this(Arrays.asList(variables));
    }

    /**
     * Construct PartialVariableCollection from a collection of variables.
     *
     * Due to limitations in MATLAB, variadic functions are not an option.  Instead create this function to take a collection
     * and construct PartialVariableCollections from it.
     * @param variables The variables to store in this collection.
     */
    public PartialVariableCollection(Collection <Variable> variables) {
        Map<String, Variable> variablesMap = new LinkedHashMap<String, Variable>(variables.size());

        final Set<EnumerationType> enumerationTypes = new LinkedHashSet<EnumerationType>();
        for (Variable var : variables) {
            variablesMap.put(var.name(), var);
            if (var.type() instanceof EnumerationType) {
                enumerationTypes.add((EnumerationType) var.type());
            }
        }

        m_variables = Collections.unmodifiableMap(variablesMap);
        m_enumerationTypes = Collections.unmodifiableSet(enumerationTypes);
    }

    public final Map<String,Variable> m_variables;
    public final Set<EnumerationType> m_enumerationTypes;
}
