/*
 * Copyright (C) 2013 Matthew Dawson <matthew@mjdsystems.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.expression;

/**
 * Provides a mechanism to request an expression resolve to a version of a given type, without fully specifying it.
 * <p>
 * When looking for the appropriate type for an expression, sometimes an applicable type has a large (or infinite)
 * potential forms of a given class.  Type markers allow for a given class to be requested without fully specifying the
 * type.  Due to this, a specific concrete specification of a type is always a marker for itself, since each type is
 * in a given type containing only itself.
 * <p>
 * The type marker concept becomes more useful when considering types like fixed point, which can have any number of
 * digits representing its type.  However, the requesting code often doesn't care how many digits are used, it just
 * wants fixed point to be considered valid.  The fixed point type can provide a special type marker that acts as a
 * marker for all valid fixed point representations.
 *
 * @author Matthew Dawson
 */
public interface TypeMarker {
    /**
     * Queries whether the given type is in the class represented by this type marker, and thus marked.
     * <p>
     * Implementations should return true if the given {@link Type} is in the class
     * of types represented by this marker, and is thus marked.  Returns false otherwise.  Any valid implementation of
     * {@link Type} can be passed in, including from other libraries.
     * @param type The type being checked.
     * @return Whether type is in the class represented by this marker.
     */
    boolean isMarkerFor(Type type);
}
